import numpy as np
import ctypes
from six import moves
from .sobol_lib2 import *  # longer sobol sequence !
from .sobol import *      
from .halton import *    

# From: https://github.com/PhaethonPrime/hammersley/blob/master/hammersley/sequences.py

# this list of primes allows up to a size 120 vector
saved_primes = [2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
                59,61,67,71,73,79,83,89,97,101,103,107,109,113,
                127,131,137,139,149,151,157,163,167,173,179,181,
                191,193,197,199,211,223,227,229,233,239,241,251,
                257,263,269,271,277,281,283,293,307,311,313,317,
                331,337,347,349,353,359,367,373,379,383,389,397,
                401,409,419,421,431,433,439,443,449,457,461,463,
                467,479,487,491,499,503,509,521,523,541,547,557,
                563,569,571,577,587,593,599,601,607,613,617,619,
                631,641,643,647,653,659]


def get_phi(p, k):
    p_ = p
    k_ = k
    phi = 0
    while k_ > 0:
        a = k_ % p
        phi += a / p_
        k_ = int(k_ / p)
        p_ *= p
    return phi


# uniform numbers' generation
def generate_uniform(n_dims=2, n_points=10, seed=123):
    np.random.seed(seed=seed)
    return np.random.random((n_dims, n_points))


# hammersley numbers' generation
def generate_hammersley(
    n_dims=2, n_points=100, primes=None
):
    def func_hammersley(
        n_dims=n_dims,
        n_points=(n_points + 1),
        primes=primes,
    ):
        primes = (
            primes if primes is not None else saved_primes
        )
        for k in moves.range(n_points):
            points = [k / n_points] + [
                get_phi(primes[d], k)
                for d in moves.range(n_dims - 1)
            ]
            yield points

    return np.array(list(func_hammersley()))[
        1 : (n_points + 1), :
    ].transpose()


# halton numbers' generation
def generate_halton(n_dims=2, n_points=10, primes=None):
    def func_halton(
        n_dims=n_dims,
        n_points=(n_points + 1),
        primes=primes,
    ):
        primes = (
            primes if primes is not None else saved_primes
        )
        for k in moves.range(n_points):
            points = [
                get_phi(primes[d], k)
                for d in moves.range(n_dims)
            ]
            yield points

    return np.array(list(func_halton()))[
        1 : (n_points + 1), :
    ].transpose()


def generate_halton_cpp(n_dims=2, n_points=10):
    try:
        n_p = n_dims*n_points
        x = halton_sequence(1, n_points, n_dims)
        xx = (ctypes.c_double * n_p).from_address(int(x))
        return np.transpose(np.array(list(xx)).reshape(n_points, n_dims))
    except:
        raise ValueError('_halton.so is not imported.')

    
# sobol numbers' generation (cpp)
def generate_sobol(n_dims=2, n_points=10):
    try:
        n_p = n_points*n_dims
        x = i8_sobol_generate(n_dims, n_points, 1)
        xx = (ctypes.c_double * n_p).from_address(int(x))
        return np.transpose(np.array(list(xx)).reshape(n_points, n_dims))
    except:
        raise ValueError('_sobol.so is not imported.')


# sobol numbers' generation
def generate_sobol2(n_dims=2, n_points=10):
    return np.array(
        i4_sobol_generate(
            m=n_dims, n=n_points, skip=2
        )
    )


# sobol numbers' generation
# def generate_sobol2(n_dims=2, n_points=10):
#    return np.array(
#        sobol_lib.i4_sobol_generate(
#            m=n_dims, n=n_points, skip=2
#        )
#    )




from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")


# Import the low-level C/C++ module
#if __package__ or "." in __name__:
    
    try:
        
        from ._sobol import *
            
    except:        
            
        import _sobol
        
    else:
        
        try:
            import builtins as __builtin__
        except ImportError:
            import __builtin__
        
        def _swig_repr(self):
            try:
                strthis = "proxy of " + self.this.__repr__()
            except __builtin__.Exception:
                strthis = ""
            return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)
        
        
        def _swig_setattr_nondynamic_instance_variable(set):
            def set_instance_attr(self, name, value):
                if name == "thisown":
                    self.this.own(value)
                elif name == "this":
                    set(self, name, value)
                elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
                    set(self, name, value)
                else:
                    raise AttributeError("You cannot add instance attributes to %s" % self)
            return set_instance_attr
        
        
        def _swig_setattr_nondynamic_class_variable(set):
            def set_class_attr(cls, name, value):
                if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
                    set(cls, name, value)
                else:
                    raise AttributeError("You cannot add class attributes to %s" % cls)
            return set_class_attr
        
        
        def _swig_add_metaclass(metaclass):
            """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
            def wrapper(cls):
                return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
            return wrapper
        
        
        class _SwigNonDynamicMeta(type):
            """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
            __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)
        
        
        
        def i8_bit_hi1(n):
            return _sobol.i8_bit_hi1(n)
        
        def i8_bit_lo0(n):
            return _sobol.i8_bit_lo0(n)
        
        def i8_max(i1, i2):
            return _sobol.i8_max(i1, i2)
        
        def i8_min(i1, i2):
            return _sobol.i8_min(i1, i2)
        
        def i8_sobol(dim_num, seed, quasi):
            return _sobol.i8_sobol(dim_num, seed, quasi)
        
        def i8_sobol_generate(m, n, skip):
            return _sobol.i8_sobol_generate(m, n, skip)
        
        def i8_uniform(b, c, seed):
            return _sobol.i8_uniform(b, c, seed)
        
        def r8_abs(x):
            return _sobol.r8_abs(x)
        
        def r8_nint(x):
            return _sobol.r8_nint(x)
        
        def r8_uniform_01(seed):
            return _sobol.r8_uniform_01(seed)
        
        def tau_sobol(dim_num):
            return _sobol.tau_sobol(dim_num)
        
        def timestamp():
            return _sobol.timestamp()
        
        



# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# check if x is int
def is_int(x):
    try:
        return int(x) == x
    except:
        return False


# check if x is float
def is_float(x):
    return isinstance(x, float)


# check if the response contains only integers
def is_factor(y):
    return all((is_int(item)) & (is_float(item)==False) for item in y)
    

# flatten list of lists
flatten = lambda l: [
    item for sublist in l for item in sublist
]




from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")


# Import the low-level C/C++ module
#if __package__ or "." in __name__:
    
    try:
        
        from ._halton import *
    
    except:
        
        import _halton
        
    else:
        
        try:
            import builtins as __builtin__
        except ImportError:
            import __builtin__
        
        def _swig_repr(self):
            try:
                strthis = "proxy of " + self.this.__repr__()
            except __builtin__.Exception:
                strthis = ""
            return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)
        
        
        def _swig_setattr_nondynamic_instance_variable(set):
            def set_instance_attr(self, name, value):
                if name == "thisown":
                    self.this.own(value)
                elif name == "this":
                    set(self, name, value)
                elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
                    set(self, name, value)
                else:
                    raise AttributeError("You cannot add instance attributes to %s" % self)
            return set_instance_attr
        
        
        def _swig_setattr_nondynamic_class_variable(set):
            def set_class_attr(cls, name, value):
                if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
                    set(cls, name, value)
                else:
                    raise AttributeError("You cannot add class attributes to %s" % cls)
            return set_class_attr
        
        
        def _swig_add_metaclass(metaclass):
            """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
            def wrapper(cls):
                return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
            return wrapper
        
        
        class _SwigNonDynamicMeta(type):
            """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
            __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)
        
        
        
        def halton(i, m):
            return _halton.halton(i, m)
        
        def halton_base(i, m, b):
            return _halton.halton_base(i, m, b)
        
        def halton_inverse(r, m):
            return _halton.halton_inverse(r, m)
        
        def halton_sequence(i1, i2, m):
            return _halton.halton_sequence(i1, i2, m)
        
        def i4vec_sum(n, a):
            return _halton.i4vec_sum(n, a)
        
        def prime(n):
            return _halton.prime(n)
        
        def r8_mod(x, y):
            return _halton.r8_mod(x, y)
        
        def timestamp():
            return _halton.timestamp()
        
        



from .nodesimulation import (    
    generate_sobol2,
    generate_uniform,
    generate_hammersley,
    generate_halton,
    
)

from .rowsubsampling import subsample


try:
    
    from .nodesimulation import generate_sobol
    from .nodesimulation import generate_halton_cpp
    
    __all__ = [
    "generate_sobol",
    "generate_sobol2",
    "generate_uniform",
    "generate_hammersley",
    "generate_halton",
    "generate_halton_cpp",
    "subsample"]

except:
    
    __all__ = [
    "generate_sobol2",
    "generate_uniform",
    "generate_hammersley",
    "generate_halton",
    "subsample",]

